"""Base for Records.
"""

from typing import (
    Protocol,
    Union,
    runtime_checkable,
    Iterable,
    Any)
from adgtk.common import FactoryBlueprint

# ----------------------------------------------------------------------
# Types and Protocols
# ----------------------------------------------------------------------


@runtime_checkable
class PresentableRecord(Protocol):
    """A record that is presentable"""

    def __str__(self) -> str:
        """Generates a string representation

        :return: A string representation of the Record 
        :rtype: str
        """

    def create_copy_of_data(self) -> dict:
        """Creates a copy of the internal data for this record. this
        protects against accidental updating of the data if needed to
        manipulate the values or keys.

        :return: a deep copy of the data
        :rtype: dict
        """

    def copy(self):
        """provides a mapping to __copy__ method. Main benefit is
        readability. no other processing occurs in this method.

        :return: a copy of the record
        :rtype: DataRecord
        """


@runtime_checkable
class SupportsFiltering(Protocol):
    """Provides a consistent definition of a filter for use across modules."""
    blueprint: FactoryBlueprint

    def is_included(self, a: Any) -> bool:
        """Filters a single objec. Intended to use as part of filtering
        an iterable object such a list or data store. Implementations
        can opt to use this when they need to define via Factory

        :param record: _description_
        :type record: Any
        :return: _description_
        :rtype: bool
        """


@runtime_checkable
class PresentableGroup(Protocol):
    """One or more PresentableRecords that act together."""
    blueprint: FactoryBlueprint
    records: Iterable[PresentableRecord]
    metadata: dict

    def __len__(self) -> int:
        """The number of records in the group"""

    def __getitem__(self, index: Union[int, slice]):
        """Get an item or a slice

        :param index: The index or slice
        :type index: Union[int, slice]
        """

    def add_record(self, record: PresentableRecord) -> None:
        """Adds a record to the group

        :param record: _description_
        :type record: PresentableRecord
        """

    def create_copy_of_data(self) -> dict:
        """Creates a copy of the internal data for this record. this
        protects against accidental updating of the data if needed to
        manipulate the values or keys.

        :return: a deep copy of the data
        :rtype: dict
        """

    def copy(self):
        """provides a mapping to __copy__ method. Main benefit is
        readability. no other processing occurs in this method.

        :return: a copy of the record group
        :rtype: DataRecordGroup
        """



"""Provides a common place for default values. Helps to avoid circular
importing."""
from typing import Literal

# Used for creating the project.toml/yaml file
# is also used when failing to load settings.

DEFAULT_FILE_FORMAT: Literal["yaml", "toml"] = "yaml"

# not changable by the user
DEFAULT_JOURNAL_REPORTS_DIR = "reports"

# changeable by the user
DEFAULT_EXP_DEF_DIR = "experiment-def"
DEFAULT_DATA_DIR = "data"
DEFAULT_SETTINGS = {
    "experiment": {
        "data_dir": DEFAULT_DATA_DIR,
        "tensorboard_dir": "runs",
        "results_dir": "results",
        "definition_dir": DEFAULT_EXP_DEF_DIR
    },
    "user_modules": [
        'agent',
        'structure',
        'environment',        
        'generation',
        "instrumentation",
        "plugin",
        'processing',
        'scenario',
        'reward',
        'policy'],
    "default_file_format": DEFAULT_FILE_FORMAT,
    "blueprint_dir": "blueprints",
    "logging": {
        "log_dir": "logs",
        "level": "basic"
    },
    "server": {
        "port": 8000,
        "host": "0.0.0.0",
        "proto": "http"
    }
}



"""Welcome to ADGTK (Agentic Data Generation Tool Kit). 

This is a pre-release development version of the project.
"""


__version__ = "0.1.0a1"

__author__ = "Fred Diehl"

__maintainer__ = "Fred Diehl"

from .common import ArgumentSetting, ArgumentType, FactoryBlueprint


"""Scenario base

Versions:
v 0.1
- mvp

References:
-

TODO:

1.0

Defects:

1.0
"""

from typing import Protocol, runtime_checkable

SCENARIO_GROUP_LABEL = "scenario"


@runtime_checkable
class Scenario(Protocol):
    """The Scenario Protocol"""

    def execute(self, name: str) -> None:
        """Execute the scenario.

        :param name: The name of the experiment (for reporting, etc)
        :type name: str
        """



"""Logging functions
"""

import logging
import sys
import os
from adgtk import __version__ as adgtk_ver
from .settings import load_settings


# ----------------------------------------------------------------------
# Future move into settings file
# ----------------------------------------------------------------------
LOG_TO_FILE = True      # future set/override perhaps?


def start_logging(
    name: str,
    surpress_chatty: bool = True,
    preview: bool = False
) -> None:
    """Sets up and starts logging for an experiment

    :param name: The name of the experiment
    :type name: str
    :param surpress_chatty: removes chatty modules httpx and openai and
        sets their log level to logging.WARNING, defaults to True
    :type surpress_chatty: bool, optional
    :param preview: logs preview only, defaults to False
    :type preview: bool, optional
    """
    settings = load_settings()

    # Setup the log directory if it doesn't already exist
    os.makedirs(settings.logging["log_dir"], exist_ok=True)

    # Setup format
    if settings.logging["level"] == "basic":
        fmt_string = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    else:
        fmt_string = "%(asctime)s %(levelname)s"\
            " %(message)s [%(module)s %(funcName)s %(lineno)d]"
        
    # cleanup name
    name = name.lower()
    if name.endswith(".toml") or name.endswith(".yaml"):
        name = name[:-5]

    log_filename = name    
    if not log_filename.endswith(".log"):
        log_filename += ".log"

    log_file = os.path.join(settings.logging["log_dir"], log_filename)
    print(f"Setting the logfile to : {log_file}")

    # Setup the logger
    if LOG_TO_FILE:
        logging.basicConfig(
            filename=log_file,
            level=logging.INFO,
            format=fmt_string)
    else:
        logging.basicConfig(
            stream=sys.stdout,
            level=logging.INFO,
            format=fmt_string)

    # now surpress logging for "chatty" modules
    if surpress_chatty:
        logging.getLogger("httpx").setLevel(logging.WARNING)
        logging.getLogger("openai").setLevel(logging.WARNING)

    if preview:
        logging.info(f"version {adgtk_ver}")
        logging.info("---------------------------------------------------")
        logging.info("----------- PREVIEW and REVIEW MODE ---------------")
        logging.info("---------------------------------------------------")
    else:
        # for when we want to log but don't expect any experiment data.
        logging.info(f"version {adgtk_ver}")
        logging.info("---------------------------------------------------")
        logging.info("------------------- NEW RUN -----------------------")
        logging.info("---------------------------------------------------")



"""Tracks various performance statistics and writes to disk"""

import os
from typing import Union, Any
import pandas as pd
from adgtk.common import FolderManager
from .base import MetricTracker


class PerformanceTracker():
    """Used to track and report on running performance. Examples include
    success rate, average time to complete a task, etc. The design is to
    be as flexible as possible to support any future user needs."""

    def __init__(self, experiment_name:str, component:str, last_only:bool=True):
        super().__init__()
        self.last_only = last_only
        self.metric_tracker = MetricTracker()
        folders = FolderManager(experiment_name)
        self.filename = os.path.join(folders.performance, f"{component}.csv")


    def register_statistic(self, label:str):
        self.metric_tracker.register_metric(label=label)

    def add_data(self, label:str, value:float|int):
        self.metric_tracker.add_data(label=label, value=value)
        
    def save_data(self):
        labels = self.metric_tracker.metric_labels()
        out_data = {}
        for label in labels:
            if self.last_only:
                value = self.metric_tracker.get_latest_value(label)
                data = [value]
            else:
                data = self.metric_tracker.get_all_data(label)                

            # now set the data
            out_data[label] = data

        df = pd.DataFrame(out_data)
        df.to_csv(self.filename)


"""Exceptions for Scenarios"""


class InvalidScenarioState(Exception):
    """Used for any scenario state that is invalid"""

    default_msg = "Invalid Scenario state."

    def __init__(self, message: str = default_msg):
        super().__init__(message)


class InsufficientData(Exception):
    """Used when there is insufficient data to perform an operation"""

    default_msg = "Insufficient data."

    def __init__(self, message: str = default_msg):
        super().__init__(message)



"""Summary goes here.

Versions:
v 0.1
- mvp

References:
-

TODO:

1.0 Consider moving to the factory module?

Defects:

1.0

Test
py -m pytest -s test/factory/test_component_factory.py
"""

from __future__ import annotations
import logging
import os
import sys
import inspect
import toml
import yaml
from typing import Any, Union, List, TYPE_CHECKING
from adgtk.journals import ExperimentJournal
from adgtk.common import FactoryBlueprint, ComponentDef, InvalidScenarioState
from adgtk.common import FolderManager
from adgtk.utils import create_line, load_settings
from .base import DuplicateFactoryRegistration, FactoryImplementable


# ----------------------------------------------------------------------
# Module Options
# ----------------------------------------------------------------------
# used for development and troubleshooting.
LOG_FACTORY_CREATE = True

README_TEXT = """BLUEPRINT_DIR
=============
This folder and its contents are re-created everytime the factory is loaded.

IMPORTANT: Do not edit any file in this folder as any changes will be lost.

The purpose of this folder is to provide you with all the blueprints if you want
to craft your experiment by hand. You can of course always use:

adgtk-mgr -b 

which will provide you an interactive experience for crafting your experiment.

Note: The current version only supports blueprints to toml. Future versions may support additional versions.
"""


def uses_factory_on_init(component: FactoryImplementable) -> bool:
    if inspect.isclass(component):
        args = inspect.signature(component).parameters
        if "factory" in args:
            return True

    return False


def uses_journal_on_init(component: FactoryImplementable) -> bool:
    if inspect.isclass(component):
        args = inspect.signature(component).parameters
        if "journal" in args:
            return True

    return False


# ----------------------------------------------------------------------
# Factory
# ----------------------------------------------------------------------


class ObjectFactory:
    """A dynamic factory that creates and manages groups and types"""

    def __init__(
        self,
        journal: ExperimentJournal,
        factory_name: str = "Object Factory",
        settings_file_override: Union[str, None] = None,
        create_blueprint_files: bool = False
    ) -> None:
        self.factory_name = factory_name
        # A common place to establish the folder manager
        # should be set before creation. So it is set at experiment run.
        # this way an Agent if designed can run different experiments
        # in parallel or series.
        self.settings_file_override = settings_file_override
        self.folder_manager: FolderManager

        self.create_blueprint_files = create_blueprint_files
        self._journal = journal

        self._registry: dict[str, dict[str, FactoryImplementable]] = {}
        self.registered_count = 0

        self.settings = load_settings()
        # formatting
        self.file_format = self.settings.default_file_format
        self.blueprint_dir = self.settings.blueprint_dir

        # Blueprint dir preperation
        if create_blueprint_files:
            if not os.path.exists(self.blueprint_dir):
                os.makedirs(self.blueprint_dir)

            readme_file = os.path.join(self.blueprint_dir, "README.md")
            with open(file=readme_file, mode="w", encoding="utf-8") as outfile:
                outfile.write(README_TEXT)

    def __len__(self) -> int:
        return self.registered_count

    def update_folder_manager(self, name: str) -> None:
        self.folder_manager = FolderManager(
            name=name,
            settings_file_override=self.settings_file_override)

    def __str__(self) -> str:
        title = "Object Factory report"
        report = ""
        report += f"{title}\n"
        report += "---------------------\n"
        for key, group in sorted(self._registry.items()):
            report += f"Group-label: {key}\n"
            for item, _ in sorted(group.items()):
                report += f"  - type: {item}\n"

        return report

    def group_report(self, group_label: str) -> str:
        """Creates a report string for a group. Primary use is in the
        command line interface.

        :param group_label: The group to filter on
        :type group_label: str
        :return: a report showing the group members
        :rtype: str
        """

        title = f"Object Factory report for group1: {group_label}"

        line = create_line(title, "=")
        report = title
        report += f"\n{line}\n"
        if group_label not in self._registry:
            report = f"ERROR: No group {group_label} found\n\n"
            report += "Valid groups are:\n"
            report += create_line("", char=".", modified=17)
            report += "\n"
            for key in self._registry.keys():
                report += f"  - {key}\n"
            return report

        for item, _ in sorted(self._registry[group_label].items()):
            report += f"  - {item:<17}  | "
            desc = self.get_description(
                group_label=group_label, type_label=item)
            report += f"{desc}\n"

        return report

    def get_description(self, group_label: str, type_label: str) -> str:
        """Gets the description of a component that is registed.

        :param group_label: The group the component is a member of
        :type group_label: str
        :param type_label: the member in the group
        :type type_label: str
        :return: a string of the description as defined in the class.
        :rtype: str
        """
        try:
            return self._registry[group_label][type_label].description
        except IndexError:
            return "ERROR. group:type not found"
        except AttributeError:
            # ideally should not happen but if it does keep going!
            return "description not set"

    def _save_blueprint(self, blueprint: FactoryBlueprint) -> None:
        file_w_path = os.path.join(
            self.blueprint_dir,
            f"{blueprint['group_label']}.{blueprint['type_label']}")

        file_w_path += ".toml"
        with open(file=file_w_path, mode="w", encoding="utf-8") as outfile:
            toml.dump(blueprint, outfile)

        return None

        # issues with yaml safe_dump and blueprints. Future work may
        # research alternative options for writing yaml to disk for the
        # blueprint. For now, defaulting to toml only. This is also an
        # edge case. Most users will hopefully use the CLI wizard. This
        # option may therefore be removed in future releases.
        if self.file_format == "toml":
            file_w_path += ".toml"
            with open(file=file_w_path, mode="w", encoding="utf-8") as outfile:
                toml.dump(blueprint, outfile)
        elif self.file_format == "yaml":
            file_w_path += ".yaml"
            with open(file=file_w_path, mode="w", encoding="utf-8") as outfile:
                try:
                    yaml.safe_dump(blueprint,
                                   outfile,
                                   default_flow_style=False,
                                   sort_keys=False)
                except yaml.representer.RepresenterError:
                    print(f"Failed to save: {blueprint['type_label']}")

    def register(
        self,
        creator: Any,
        group_label_override: Union[str, None] = None,
        type_label_override: Union[str, None] = None,
        log_entry: bool = True
    ) -> None:
        """Register an object.

        Args:
        :param group_label_overide: The group to which it belongs
        :type group_label_overide: str
        :param type_label_overide: the type within that group
        :type type_label_overide: str
        :param creator: The class to create
        :type creator: Any (implements Protocol FactoryImplementable)
        :raises DuplicateFactoryRegistration: The type label exists
        :raises TypeError: creator does not implement protocol
        """

        group_label: str = creator.blueprint["group_label"]
        type_label: str = creator.blueprint["type_label"]

        if group_label_override is not None:
            group_label = group_label_override
        if type_label_override is not None:
            type_label = type_label_override

        # Safety check
        if not isinstance(creator, FactoryImplementable):
            msg = f"Creator {creator} does not Implement FactoryImplementable"

            raise TypeError(msg)
        if group_label not in self._registry:
            self._registry[group_label] = {}

        # get groups
        c_group = self._registry[group_label]

        # and add to groups
        if type_label in c_group:
            raise DuplicateFactoryRegistration

        c_group[type_label] = creator

        # and create a local file should the user want to create by hand
        if self.create_blueprint_files:
            self._save_blueprint(creator.blueprint)

        self.registered_count += 1

    def unregister(self, group_label: str, type_label: str) -> None:
        """Unregisters a type from a group

        :param group_label: The group to which it belongs
        :type group_label: str
        :param type_label: the type within that group
        :type type_label: str
        :raises KeyError: Group label not found
        """
        if group_label in self._registry:
            c_group = self._registry[group_label]
            if type_label in c_group:
                c_group.pop(type_label, None)
        else:
            raise KeyError("Group label not found")

        self.registered_count -= 1

    def get_blueprint(
        self,
        group_label: str,
        type_label: str
    ) -> FactoryBlueprint:
        """Gets a blueprint for a component creator

        :param group_label: the group label
        :type group_label: str
        :param type_label: The type label
        :type type_label: str
        :raises KeyError: Group not found
        :raises KeyError: Type not found
        :return: The component blueprint
        :rtype: FactoryBlueprint
        """

        if group_label in self._registry:
            c_group = self._registry[group_label]
        else:
            raise KeyError("Group label not found")

        if type_label in c_group:
            component = c_group[type_label]
            return component.blueprint

        raise KeyError("Type label not found")

    def create(self, component_def: ComponentDef) -> Any:
        """creates a component based on a blueprint

        :param component_def: The item to build
        :type component_def: ComponentDef
        :raises KeyError: unable to find the object
        :return: A created object using the blueprint
        :rtype: Any
        """

        try:
            if "group_label" not in component_def:
                msg = "Invalid component definition, missing group_label"
                raise KeyError(msg)

            if "type_label" not in component_def:
                msg = "Invalid component definition, missing group_label"
                raise KeyError(msg)

            if component_def["group_label"] not in self._registry:
                msg = f'{component_def["group_label"]} not in factory. '\
                      "Unable to create"
                logging.error(msg)
                print(msg)
                sys.exit(1)
            g_label = component_def["group_label"]
            t_label = component_def["type_label"]
            # get the component_create
            component_create = self._registry[g_label][t_label]
        except KeyError as e:
            msg = "No valid creator found at: "\
                  f"{component_def['group_label']}:{component_def['type_label']}"
            logging.error(e)
            raise KeyError(e) from e

        # do we init w/factory & journal, factory only, or none?
        factory_flag = uses_factory_on_init(component_create)
        j_flag = uses_journal_on_init(component_create)

        if factory_flag and j_flag:
            if LOG_FACTORY_CREATE:
                g_label = component_def["group_label"]
                t_label = component_def["type_label"]
                msg = f"creating {g_label}:{t_label} with Journal and Factory"
                logging.info(msg)
            try:
                if inspect.isclass(component_create):
                    new_component = component_create(
                        factory=self,
                        journal=self._journal,
                        **component_def["arguments"])
                else:
                    msg = f"invalid type in factory for {g_label}|{t_label}"
                    logging.error(msg)
                    raise InvalidScenarioState(msg)
            except TypeError as e:
                msg = f"Invalid settings for {g_label}|{t_label}. "\
                    "Unable to create using the factory. See log for "\
                    "more details."
                logging.error(msg)
                logging.error(e)
                raise TypeError(msg) from e

            return new_component

        elif factory_flag:
            if LOG_FACTORY_CREATE:
                g_label = component_def["group_label"]
                t_label = component_def["type_label"]
                msg = f"creating {g_label}:{t_label} with Factory, no Journal"
                logging.info(msg)
            try:
                if inspect.isclass(component_create):
                    new_component = component_create(
                        factory=self,
                        **component_def["arguments"])
                else:
                    msg = f"invalid type in factory for {g_label}|{t_label}"
                    logging.error(msg)
                    raise InvalidScenarioState(msg)

            except TypeError as e:
                msg = f"Invalid settings for {g_label}|{t_label}. "\
                    "Unable to create using the factory. See log for "\
                    "more details."
                logging.error(msg)
                logging.error(e)
                raise TypeError(msg) from e

            return new_component

        elif j_flag:
            if LOG_FACTORY_CREATE:
                g_label = component_def["group_label"]
                t_label = component_def["type_label"]
                msg = f"creating {g_label}:{t_label} with Journal, no Factory"
                logging.info(msg)

            try:
                if inspect.isclass(component_create):
                    new_component = component_create(
                        journal=self._journal,
                        **component_def["arguments"])
                else:
                    msg = f"invalid type in factory for {g_label}|{t_label}"
                    logging.error(msg)
                    raise InvalidScenarioState(msg)

            except TypeError as e:
                msg = f"Invalid settings for {g_label}|{t_label}. "\
                    "Unable to create using the factory. See log for "\
                    "more details."
                logging.error(msg)
                logging.error(e)
                raise TypeError(msg) from e

            return new_component

        else:
            if LOG_FACTORY_CREATE:
                g_label = component_def["group_label"]
                t_label = component_def["type_label"]
                msg = f"creating {g_label}:{t_label} w/out Factory or journal"
                logging.info(msg)

            try:
                if inspect.isclass(component_create):
                    new_component = component_create(
                        **component_def["arguments"])
                else:
                    msg = f"invalid type in factory for {g_label}|{t_label}"
                    logging.error(msg)
                    raise InvalidScenarioState(msg)
            except TypeError as e:
                msg = f"Invalid settings for {g_label}|{t_label}. "\
                    "Unable to create using the factory. See log for "\
                    "more details."
                logging.error(msg)
                logging.error(e)
                raise TypeError(msg) from e

            return new_component

    def registry_listing(
        self,
        group_label: Union[str, None] = None
    ) -> List[str]:
        """Lists factory entries. if no group listed it returns groups,
        else if a group is entered it returns all types in that group.

        :param group_label: A group label, defaults to None
        :type group_label: Union[str, None], optional
        :return: A listing of types in a group or all groups
        :rtype: List[str]
        """        
        if group_label is None:
            return list(self._registry.keys())
        else:
            try:           
                group = self._registry[group_label]                
                return list(group.keys())
            except KeyError as e:
                msg = f"Invalid group_label: {group_label} not found."
                raise KeyError(msg) from e



"""Internal tracking. Provides useful data structures.
"""
import logging
import os
import copy
from typing import Union, Iterable
from adgtk.utils import plot_single_line
from adgtk.common.exceptions import InsufficientData
# ----------------------------------------------------------------------
# Constants
# ----------------------------------------------------------------------
DEBUG_TO_CONSOLE = False

# ----------------------------------------------------------------------
# Tracking of data
# ----------------------------------------------------------------------


class MetricTracker():
    """Used for tracking metrics."""

    def __init__(self):
        self.metrics = {}
        self.metadata = {}

    def register_metric(
        self,
        label: str,
        metadata: Union[dict, None] = None
    ) -> bool:
        """Registers a metric.

        :param label: The label of the metric
        :type label: str
        :param metadata: _description_
        :type metadata: Union[dict, None]
        :return: T: created, F: did not create
        :rtype: bool
        """

        if metadata is not None:
            if label not in self.metadata:
                self.metadata[label] = metadata
        else:
            if label not in self.metadata:
                self.metadata[label] = {}

        if label not in self.metrics:
            self.metrics[label] = []
            return True
        return False

    def add_raw_data(self, label: str, values: Iterable) -> None:
        """Adds data as-is by iterating through and adding one by one.

        :param label: The label of the metric
        :type label: str
        :param values: the data to add
        :type values: Iterable
        """
        for data in values:
            self.add_data(label=label, value=data)

    def add_data(self, label: str, value: Union[int, float]) -> None:
        """Adds data

        :param label: The label of the metric
        :type label: str
        :raises KeyError: Label is not found
        :param value: the data to add
        :type Union[int, float]
        """
        if label not in self.metrics:
            raise KeyError("Invalid metric")

        self.metrics[label].append(value)

        if DEBUG_TO_CONSOLE:
            print(f"MetricTracker adding {value} to {label}")
            print(f"Updated Metrics: {self.metrics[label]}")

    # TODO: Consider moving from bool to raise Exception when an
    # entry exists. not needed for MVP.
    def metric_exists(self, label: str) -> bool:
        """Does a metric exist?

        :param label: The label of the metric
        :type label: str
        :return: T T: exists, F: does not
        :rtype: bool
        """
        if label not in self.metrics:
            return False

        return True

    def remove_metric(self, label: str) -> None:
        """Removes a metric from being tracked

        :param label: The label of the metric to remove
        :type label: str
        """
        if label in self.metrics:
            del self.metrics[label]

        if label in self.metadata:
            del self.metadata[label]

    def metric_labels(self) -> list:
        """Gets a list of metrics currently tracking

        :return: a list of the labels
        :rtype: list
        """
        return list(self.metrics.keys())

    def get_latest_value(self, label: str) -> float:
        """Gets the latest value from a label

        :param label: The label of the metric to get latest value
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the latest value
        :rtype: float
        """
        if label not in self.metrics:
            msg = f"Requested invalid label: {label}"
            logging.error(msg)
            raise KeyError("Invalid metric")
        elif len(self.metrics[label]) == 0:
            return 0
        else:
            return self.metrics[label][-1]

    def get_average(self, label: str) -> float:
        """Returns the average of all stored values for the label

        :param label: The label of the metric to get avg of
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the average value
        :rtype: float
        """
        if label not in self.metrics:
            if DEBUG_TO_CONSOLE:
                print(f"METRIC_TRACKER_DATA: {self.metrics}")
                msg = f"Requested invalid label: {label}"
                print(msg)
            logging.error(msg)
            raise KeyError("Invalid metric")
        elif len(self.metrics[label]) == 0:
            return 0
        else:
            return sum(self.metrics[label]) / len(self.metrics[label])

    def get_sum(self, label: str) -> float:
        """Returns the sum of all stored values for the label

        :param label: The label of the metric to get sum of
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the sum of all stored values
        :rtype: float
        """
        if label not in self.metrics:
            msg = f"Requested invalid label: {label}"
            logging.error(msg)
            raise KeyError("Invalid metric")
        elif len(self.metrics[label]) == 0:
            return 0
        else:
            return sum(self.metrics[label])

    def clear_metric(self, label: str) -> None:
        """Clears the values of a  metric

        :param label: The label of the metric to clear data from
        :type label: str
        """
        self.metrics[label] = []

    def reset(self) -> None:
        """Deletes all data and labels and resets to no metrics tracked.
        """
        self.metrics = {}

    def measurement_count(self, label: str) -> int:
        """Returns the count of observations for a metric


        :param label: The label of the metric to get count of
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the count of all entries
        :rtype: int
        """
        if label not in self.metrics:
            raise KeyError("Invalid metric")

        return len(self.metrics[label])

    def get_all_data(self, label: str) -> list:
        """Gets all data for a metric.

        :param label: The label of the metric to get data from
        :type label: str
        :raises KeyError: Invalid Metric
        :return: the data as a list
        :rtype: list
        """
        if self.metric_exists(label):
            return copy.deepcopy(self.metrics[label])

        msg = f"Requested invalid label: {label}"
        logging.error(msg)
        raise KeyError("Invalid metric")

    def get_metadata(self, label: str) -> dict:
        if label in self.metadata:
            return copy.deepcopy(self.metadata[label])

        msg = f"Requested invalid metadata for label: {label}"
        logging.error(msg)
        return {}

    # ------------------------ Plotting --------------------------------

    def line_plot(self, label: str, folder: str, file_prefix: str) -> str:
        """MVP Line plot. Does a basic line plot using the data tracked.

        :param label: The metric to plot
        :type label: str
        :param folder: The location to save the file
        :type folder: str
        :param file_prefix: The prefix of the filename. ex engine name
        :type file_prefix: str
        :return True if the file was created
        :rtype bool
        """

        file_w_path = os.path.join(folder, f"{file_prefix}.{label}.png")
        data = self.get_all_data(label=label)
        if len(data) > 0:
            result = plot_single_line(
                data=data,
                filename=file_w_path,
                title=label)

            if result:
                return file_w_path

        raise InsufficientData



"""Common module
"""
from .exceptions import InvalidScenarioState, InsufficientData
from .structure import (
    ExperimentDefinition,
    ComponentDef,
    ArgumentType,
    ArgumentSetting,
    FactoryBlueprint,
    is_blueprint,
    is_valid_arg_setting,
    default_is_arg_type,
    convert_exp_def_to_string,
    build_tree)

from .results import (
    METRICS_DATA_FOLDER,
    METRICS_FOLDER,
    METRICS_IMG_FOLDER,
    FolderManager)

from .defaults import (
    DEFAULT_JOURNAL_REPORTS_DIR,
    DEFAULT_SETTINGS,
    DEFAULT_DATA_DIR,
    DEFAULT_EXP_DEF_DIR,
    DEFAULT_FILE_FORMAT)



"""Builds and manages experiments"""
import logging
import sys
import os
from typing import Union, cast
import toml
import yaml
from adgtk.factory import InvalidBlueprint
from adgtk.common import (
    FactoryBlueprint,
    ExperimentDefinition,
    ArgumentSetting,
    ArgumentType,
    convert_exp_def_to_string)
from adgtk.scenario import ScenarioManager, SCENARIO_GROUP_LABEL
from adgtk.utils import (
    get_user_input,
    get_more_ask,
    create_line)


class ExperimentBuilder():
    """Used to build experiments"""

    def __init__(
        self,
        experiment_definition_dir: str,
        load_user_modules:list,
        scenario_manager: Union[ScenarioManager, None] = None,
    ):
        self.experiment_definition_dir = experiment_definition_dir
        self.outer_configuring = "Scenario"     # for list processing
        self.currently_configuring = "Experiment"
        self.file_format = "yaml"
        self.last_line = ""

        if scenario_manager is None:
            self.scenario_manager = ScenarioManager(
                blueprint_dir="blueprints",
                load_user_modules=load_user_modules,
                experiment_definition_dir=experiment_definition_dir)
        else:
            self.scenario_manager = scenario_manager

    def get_registered(
        self,
        group: str,
        to_console: bool = True
    ) -> list:
        """Gets the type listing and if to_console prints to console as well

        :param manager: The scenario manager to work with
        :type manager: ScenarioManager
        :param group: The group to fetch
        :type group: str
        :param to_console: print to console?, defaults to True
        :type to_console: bool, optional
        :return: a list of valid types for this group
        :rtype: list
        """
        try:
            registered_types = self.scenario_manager.registry_listing(group)
        except KeyError:
            err_msg = "group not found. Check factory"
            if group == SCENARIO_GROUP_LABEL:                
                err_msg = "no scenarios found. Review your factory and"\
                          " code. The scenario should be listed when you "\
                          f"review the factory. observed {group}"
            else:
                if isinstance(group, str):
                    err_msg = f"group not found: {group}. Check factory"

            logging.error(err_msg)
            sys.exit(1)
        if not to_console:
            return registered_types

        group_title = f"\n Registered {group} types are :"
        line_list = ["."] * len(group_title)
        line = "".join(line_list)
        print(group_title)
        print(line)
        print()

        for label in registered_types:
            label_text = f" - {label}"
            desc = self.scenario_manager.get_description(
                group_label=group,
                type_label=label)

            print(f"{label_text:<30} | {desc}")

        print()  # extra space
        return registered_types

    def build_interactive(self, name: Union[str, None] = None):
        """Interactive building of an experiment. This is the entry
        point for the experiment builder. It provides the UX for
        creating an experiment and saving to disk.

        :param name: The name of the experiment, defaults to None
        :type name: Union[str, None], optional
        """

        exp_title = ". Experiment builder wizard ."
        self.last_line = create_line(exp_title, char=".")
        print(self.last_line)
        print(exp_title)
        print(self.last_line)

        # -------------------------------------------------------------
        # get the global settings, name, description
        # -------------------------------------------------------------
        if name is None:
            exp_name = get_user_input(
                configuring=self.currently_configuring,
                request="Please enter the name of the experiment",
                requested="str",
                allow_whitespace=False,
                helper="This should be a unique name. This is your filename")
        else:
            exp_name = name

        comments = get_user_input(
            configuring=self.currently_configuring,
            requested="str",
            request="Short description of the experiment",
            max_characters=50,
            helper="This should be less than 50 characters")

        if not isinstance(comments, str):
            comments = ""
            logging.warning("Error processing comments input")

        # start working on the scenario
        self.currently_configuring = "Scenario"
        scenario_setting = ArgumentSetting(
            help_str="\nWhat scenario do you wish to construct from?",
            group_label="scenario",
            argument_type=ArgumentType.BLUEPRINT)
        exp_conf = self._proc_blueprint(scenario_setting)
        experiment = ExperimentDefinition(
            configuration=exp_conf, comments=comments)
        
        # ensure group label is set post processing.
        exp_conf["group_label"] = SCENARIO_GROUP_LABEL
        preview = convert_exp_def_to_string(exp_conf)
        intro = f"\nExperiment preview for {exp_name}: "
        print(intro)
        print(preview)
        file_w_path = "not-set"
        if self.file_format == "toml":
            file_w_path = os.path.join(
                self.experiment_definition_dir,
                f"{exp_name}.toml")
        elif self.file_format == "yaml":
            file_w_path = os.path.join(
                self.experiment_definition_dir,
                f"{exp_name}.yaml")
        else:
            msg = f"unknown format {self.file_format}. Reverting to yaml"
            logging.error(msg)
            self.file_format = "yaml"
            file_w_path = os.path.join(
                self.experiment_definition_dir,
                f"{exp_name}.yaml")
        
        save_as = get_user_input(
            configuring=self.currently_configuring,
            request="Save file as",
            requested="str",
            default_selection=file_w_path,
            allow_whitespace=False)

        # take the default?
        if isinstance(save_as, str):
            if len(save_as) == 0:
                save_as = file_w_path

            with open(save_as, encoding="utf-8", mode="w") as outfile:
                if self.file_format == "toml":
                    toml.dump(experiment, outfile)
                elif self.file_format == "yaml":
                    yaml.safe_dump(
                        experiment,
                        outfile,
                        default_flow_style=False,
                        sort_keys=False)
        else:
            logging.error("Unable to save file due to unexpected input")

    def _proc_blueprint(self, setting: ArgumentSetting) -> FactoryBlueprint:
        """Processes a blueprint argugment

        :param setting: The setting to process
        :type setting: ArgumentSetting
        :raises InvalidBlueprint: Missing Blueprint definition
        :return: A blueprint for the factory
        :rtype: FactoryBlueprint
        """

        exp_def: FactoryBlueprint = {
            "group_label": "experiment",    # not used in this context
            "type_label" : "",
            "arguments" : {},
        }
        if "group_label" not in setting:
            msg = f"Invalid blueprint: {setting}. Missing group_label"
            logging.error(msg)
            raise InvalidBlueprint(msg)

        try:
            type_listing = self.get_registered(
                group=setting["group_label"], to_console=False)
        except ValueError:
            print(f"ERROR: group not found {setting['default_value']}")
            print(self.scenario_manager)
            sys.exit()


        if len(type_listing) == 1:

            # skip the UX for the built-in objects that have only one
            if setting["group_label"]  == "measurement-engine":
                pass
            elif setting["group_label"]  == "measurement-set":
                pass
            else:
                msg = f"\nNOTE: Factory only has option: {setting['group_label']}:"
                msg += f"{type_listing[0]}"
                self.last_line = create_line(msg, char=".")
                print(msg)
                print(self.last_line)
                print()

            
            type_label = type_listing[0]

        else:
            # re-pull, but display to console
            type_listing = self.get_registered(
                group=setting["group_label"], to_console=True)

            type_label = get_user_input(
                configuring=self.currently_configuring,
                request=setting["help_str"],
                choices=type_listing,
                requested="str")

        blueprint = self.scenario_manager.get_blueprint(
            group_label=setting["group_label"], type_label=type_label)
        
        if "introduction" in blueprint:
            print()
            print()
            self.last_line = create_line(
                text=self.last_line, char="=", title=blueprint["group_label"])
            print(self.last_line)
            print(blueprint["introduction"])
            self.last_line = create_line(text=self.last_line, char="=")
            print(self.last_line)

        exp_def["type_label"] = type_label
        if blueprint is None:
            raise InvalidBlueprint("Missing Blueprint definition")

        self.currently_configuring = \
            f"{blueprint['group_label']}:{blueprint['type_label']}"

        exp_def["arguments"] = self._proc_arguments(blueprint["arguments"])
        return exp_def

    def _proc_arguments(self, arguments: dict) -> dict:
        """Processes the arguments for the blueprint

        :param arguments: The arguments to process
        :type arguments: dict
        :return: The processed arguments
        :rtype: dict
        """

        exp_config = {}
        for key, value in arguments.items():
            exp_config[key] = self._proc_arg(setting=value)

        return exp_config

    def _proc_arg(
        self,
        setting: ArgumentSetting
    ) -> Union[list, int, str, float, bool, dict, FactoryBlueprint]:
        """Process a single argument

        :param setting: The argument to process
        :type setting: ArgumentSetting
        :raises InvalidBlueprint: Invalid blueprint found
        :return: The result based on the requested type in the setting
        :rtype: Union[list, int, str, float, bool, dict, FactoryBlueprint]
        """
        if "argument_type" not in setting:
            msg = f"Invalid blueprint: {setting}"
            logging.error(msg)
            raise InvalidBlueprint(msg)

        # cover the optional key(s)
        if "default_value" not in setting:
            setting["default_value"] = None
    

        if setting["argument_type"] == ArgumentType.ML_STRING:
            return get_user_input(
                configuring=self.currently_configuring,
                request=setting["help_str"],
                default_selection=setting["default_value"],
                requested="ml-str",
                helper="Press [Esc] followed by [Enter] to complete input")
            
        if setting["argument_type"] == ArgumentType.STRING:

            return get_user_input(
                configuring=self.currently_configuring,
                request=setting["help_str"],
                default_selection=setting["default_value"],
                requested="str",
                helper="please enter a string")
        elif setting["argument_type"] == ArgumentType.INT:
            return get_user_input(
                configuring=self.currently_configuring,
                request=setting["help_str"],
                default_selection=setting["default_value"],
                requested="int",
                helper="please enter an Integer")
        elif setting["argument_type"] == ArgumentType.FLOAT:
            return get_user_input(
                configuring=self.currently_configuring,
                request=setting["help_str"],
                default_selection=setting["default_value"],
                requested="float",
                helper="please enter a float")
        elif setting["argument_type"] == ArgumentType.BOOL:
            return self._proc_bool(setting=setting)
        elif setting["argument_type"] == ArgumentType.BLUEPRINT:
            return self._proc_blueprint(setting=setting)
        elif setting["argument_type"] == ArgumentType.LIST:
            return self._proc_list(setting=setting)
        elif setting["argument_type"] == ArgumentType.DICT:
            return self._proc_dict(setting=setting)

    def _get_bool_reply(self, help_str:str,default:Union[str, bool]) -> bool:
        if isinstance(default, str):
            default_str = default.lower()
        else:
            if default:
                default_str = "true"
            else:
                default_str = "false"

        value = get_user_input(
            request=help_str,
            requested="str",
            choices=["true", "false"],
            default_selection=default_str,
            configuring=self.currently_configuring)

        if isinstance(value, str):
            if value.lower() == "true":
                return True

        return False               

    def _proc_bool(self, setting: ArgumentSetting) -> bool:
        """Process a boolean setting request

        :param setting: The requested setting
        :type setting: ArgumentSetting
        :return: The user input
        :rtype: bool
        """
        return self._get_bool_reply(
            help_str=setting["help_str"],
            default=setting["default_value"])
        
    def _proc_list(self, setting: ArgumentSetting) -> list:
        """Process a list of entries from the user

        :param setting: The requested argument entries
        :type setting: ArgumentSetting
        :return: The user input
        :rtype: list
        """

        # Introductions
        
        if "list_intro" in setting:
            self.last_line = create_line(
                text=self.last_line, char=".", title="list entry")
            print(self.last_line)
            if "group_label" in setting:
                title = create_line(
                text=self.last_line, char="=", title=setting["group_label"])    
            else:
                title = create_line(text=self.last_line, char="=")    
            
            print(setting["list_intro"])
            print(title)
                
        if "list_arg_type" not in setting:
            msg = f"Invalid blueprint: {setting}. Missing list_arg_type"
            logging.error(msg)
            raise InvalidBlueprint(msg)
        
        arg_setting = ArgumentSetting(
                help_str=setting["help_str"],                
                argument_type=setting["list_arg_type"])
    
        # copy over optional?
        if "default_value" in setting:
                arg_setting["default_value"] = setting["default_value"]
        if "introduction" in setting:
                arg_setting["introduction"] = setting["introduction"]
        
        if setting["list_arg_type"] == ArgumentType.BLUEPRINT:
            if "list_group_label" not in setting:
                msg = f"Invalid blueprint: {setting}. Missing list_group_label"
                logging.error(msg)
                raise InvalidBlueprint(msg)
            arg_setting["group_label"] = setting["list_group_label"]  
                
        # setup
        items = []

        more = True
        print(setting['help_str'])

        # Is empty ok?
        if "list_min" in setting:
            if setting["list_min"] == 0:
                no_data = self._get_bool_reply(
                    help_str="Do you want to set as []",
                    default=True)

                if no_data:
                    return []        


        self.outer_configuring = self.currently_configuring
        while more:
            items.append(self._proc_arg(setting=arg_setting))
            more = get_more_ask(self.currently_configuring)
        print()
        self.currently_configuring = self.outer_configuring

        return items

    def _proc_dict(self, setting: ArgumentSetting) -> dict:
        """Processes collection of a dictionary key value pairs from the
        user.

        :param setting: The argument to process
        :type setting: ArgumentSetting
        :return: The user input
        :rtype: dict
        """

        data = {}
        more = True

        while more:
            key = get_user_input(
                request=f"{setting['help_str']} [key]",
                configuring=self.currently_configuring,
                requested="str",
                min_characters=1,
                allow_whitespace=False)

            value = get_user_input(
                request=f"{setting['help_str']} [{key} value]",
                configuring=self.currently_configuring,
                requested="str",
                allow_whitespace=True)

            if key in data:
                print(f"Key {key} already exists. no action taken.")
            else:
                data[key] = value

            more = get_more_ask(self.currently_configuring)

        return data



"""Common Utils"""

from .cli import get_user_input, get_more_ask, create_line, clear_screen
from .formatting import (
    get_timestamp_now,
    process_possible_yaml,
    count_prefix_spaces)
from .logs import start_logging
from .plotting import plot_single_line
from .processing import (
    string_to_bool,
    get_pair_from_iterable,
    get_sample_from_iterable)
from .settings import load_settings
from .text import camel_case_generation



# pylint: disable=pointless-string-statement
"""The primary purpose of this module is to support the built-in
components. Secondarly the framework itself can use this for processing.
The overall concept is not to fully implement a database but rather
provide a Protocol for any implementation of a wrapper for a database.

By using a consistent wrapper it enables ease of extensibility as well
as code maintenance as it allows for loose coupling across Modules.

Implementations can opt to inherit the DataStore class or implement the
prootocl directly.
"""

# pylint: disable=unused-argument
from __future__ import annotations
from enum import Enum, auto
from typing import (
    Union,
    Protocol,
    runtime_checkable,
    Any,
    Iterable)
from adgtk.components.data.records import PresentableRecord

# ----------------------------------------------------------------------
# Types and Protocols
# ----------------------------------------------------------------------


class SearchOption(Enum):
    """Search capabilities for a data store"""
    NONE = auto()


@runtime_checkable
class RecordStore(Protocol):
    """Provides a consistent way for the built-in objects to interact
    with a data store such as vector database. As most database systems
    differ in their implementations this protocol provides a wrapper
    pattern for any potential use when built-in components are used."""

    # -----------------------------------------------------------------
    # Required methods
    # -----------------------------------------------------------------

    def insert(self, record: PresentableRecord) -> None:
        """Insert a single record

        :param record: _description_
        :type record: PresentableRecord
        """

    def bulk_insert(self, records: Iterable[PresentableRecord]) -> None:
        """insert several records at once

        :param records: _description_
        :type records: Iterable[PresentableRecord]
        """

    def clear_all_records(self) -> None:
        """Clears all records from the system        
        """

    def __len__(self) -> int:
        """The number of the records in the store

        :return: The count of records
        :rtype: int
        """


@runtime_checkable
class CanFindRandomRecord(Protocol):
    """Can find random record"""

    def find_random_record(
        self,
        search_filters: Union[dict, None] = None,
        not_record: Union[PresentableRecord, None] = None,
        **kwargs
    ) -> list:
        """Finds a single record. If a search cannot use a filter
        then at most it should log the issue but not stop as long
        as it can return a record.

        Feature flag: can_get_random_record

        :param near_record: The record to compare against,
            defaults to None
        :type not_record: PresentableRecord
        :param search_filters: the search guidance
        :type search_filters: Union[dict, None]
        :return: a single random record if sufficient records exist
        :rtype: list
        """


@runtime_checkable
class CanRebuildFromDisk(Protocol):
    """Can rebuild from disk"""

    def rebuild_from_disk(self, filename: str) -> bool:
        """Replaces any current records with ones loaded from disk.

        :param filename: The filename with path
        :type filename: str
        :return: Success loading.
        :rtype: bool
        """


@runtime_checkable
class CanSaveToDisk(Protocol):
    """Can use the save_to_disk method w/filename"""

    def save_to_disk(self, filename: str) -> None:
        """Saves to disk all records. For wrapper classes this method
        should export from the database into a common format if the db
        itself does not implement this capability.

        :param filename: The filename with path
        :type filename: str
        """


@runtime_checkable
class CanShuffleRecords(Protocol):
    """Can use the shuffle method to shuffle records"""

    def shuffle(self) -> None:
        """Shuffles the order of records.
        Feature Flag: can_shuffle
        """


@runtime_checkable
class CanExportRecordsToDict(Protocol):
    """Can export records to Dict"""

    def export_to_dict(self, filters: Union[dict, None]) -> dict:
        """Exports the data based on the filters

        Feature Flag: can_export_to_dict

        :param filters: what if any should be used to filter the records
        :type filters: Union[dict, None]
        :return: a dict with requested data
        :rtype: dict
        """


@runtime_checkable
class CanImportRecordsToDict(Protocol):
    """Can import records to Dict"""

    def import_from_dict(
        self,
        data: dict,
        metadata: Union[dict[str, Any], None] = None
    ) -> bool:
        """Imports records from disk.

        Feature Flag: can_import_from_dict

        :param data: _description_
        :type data: dict
        :param metadata: _description_, defaults to None
        :type metadata: dict, optional
        :return: _description_
        :rtype: bool
        """


@runtime_checkable
class CanGetAllRecords(Protocol):
    """Can export records to Dict"""

    def get_all_records(self, as_copy: bool = True) -> list:
        """Exports all records in the store.

        feature flag: can_export_all_records_to_list

        :param as_copy: export as a new object, defaults to True
        :type as_copy: bool, optional
        :return: a list of all records
        :rtype: list
        """


@runtime_checkable
class CanSearchForSimilar(Protocol):
    """Can search for similar"""

    def search_for_similar(
        self,
        near_record: PresentableRecord,
        search_filters: Union[dict, None],
        **kwargs
    ) -> list:
        """Searches for a similar record. If a search cannot use a
        filter then at most it should log the issue but not stop as long
        as it can return one or more records.

        feature Flag: can_search_by_proximity

        :param near_record: The record to compare against
        :type near_record: PresentableRecord
        :param search_filters: the search guidance
        :type search_filters: Union[dict, None]
        :return: a list of one or more records.
        :rtype: list
        """


@runtime_checkable
class CanFindByTerm(Protocol):
    """Can find by term"""

    def find_by_term(self, term: str) -> list:
        """Searches not by record but by term.

        Feature Flag: can_search_by_term

        :param term: the term to search for
        :type term: str
        :return: a list of Records
        :rtype: list
        """



"""Common structure. Primary purpose is to avoid circular imports as
well as consistent data structure.
"""

import logging
from enum import Enum, auto
from typing import TypedDict, Required, Any, Union, cast, Optional, NotRequired
from numbers import Number
import anytree


# ----------------------------------------------------------------------
# Common structures
# ----------------------------------------------------------------------
# These structures are used for blueprint/generation of experiment
# blueprints only. They are NOT used when writing to disk the experiment
# which will be just the values.


class ArgumentType(Enum):
    """Identify the argument type."""
    BLUEPRINT = auto()
    INT = auto()
    FLOAT = auto()
    STRING = auto()
    ML_STRING = auto()
    LIST = auto()
    DICT = auto()
    BOOL = auto()


class ArgumentSetting(TypedDict):
    """The indiviudal setting."""
    help_str: Required[str]
    argument_type: Required[ArgumentType]    
    default_value: NotRequired[Any]
    group_label: NotRequired[str]               # require when Blueprint
    list_arg_type: NotRequired[ArgumentType]    # arg type for lists
    list_group_label: NotRequired[str]          # for list of Blueprints
    list_intro: NotRequired[str]                # UX for lists
    list_min: NotRequired[int]                  # minimum list count
    introduction: NotRequired[str]

    

class FactoryBlueprint(TypedDict):
    """A Blueprint is used to create a template for an experiment. The
    blueprint pattern is seeking to create a generic template for a user
    of the package to create their own specifications without needing to
    code one. A common set of factories takes these specifications and
    creates the required objects in order to perform an experiment. The
    creators (typically a Class) must include a FactoryBlueprint as well
    in order to ensure consistency from user defined "experiment" to the
    factories.

    group_label     : identifies which group. Useful for defaults.
    type_label      : the identifier within the group of the creator
    arguments       : the arguments to be passed on init to the creator
    introduction    : used to introduce a component to an agent or human
    """    
    group_label: Required[str]
    type_label: Required[str]
    arguments: Required[dict[str, ArgumentSetting]]
    introduction: NotRequired[str]


class ExperimentDefinition(TypedDict):
    """The outer wrapper of an experiment. writes to disk, etc

    :param TypedDict: _description_
    :type TypedDict: _type_
    """
    configuration: Required[FactoryBlueprint]
    comments: str


class ComponentDef(TypedDict):
    """A ComponenttDef is used to create an ojbect. The    
    group_label     : identifies which group. Useful for defaults.
    type_label      : the identifier within the group of the creator
    arguments       : the arguments to be passed on init to the creator
    """
    group_label: Required[str]
    type_label: Required[str]
    arguments: Required[dict[str, Union[str, list, int, float, dict]]]

# ----------------------------------------------------------------------
# functions
# ----------------------------------------------------------------------


def default_is_arg_type(sample: dict) -> bool:
    """Validates the default argument aligns with the type

    :param sample: the sample
    :type sample: ArgumentSetting
    :return: expected type
    :rtype: True
    """
    valid = True
    if not isinstance(sample, dict):
        valid = False
    elif sample["argument_type"] == ArgumentType.BLUEPRINT:
        valid = isinstance(sample["default_value"], str)
    elif sample["argument_type"] == ArgumentType.STRING:
        valid = isinstance(sample["default_value"], str)
    elif sample["argument_type"] == ArgumentType.INT:
        valid = isinstance(sample["default_value"], int)
    elif sample["argument_type"] == ArgumentType.FLOAT:
        valid = isinstance(sample["default_value"], float)
    elif sample["argument_type"] == ArgumentType.BOOL:
        valid = isinstance(sample["default_value"], bool)
    elif sample["argument_type"] == ArgumentType.DICT:
        valid = isinstance(sample["default_value"], dict)
    elif sample["argument_type"] == ArgumentType.LIST:
        valid = isinstance(sample["default_value"], list)
    else:
        valid = False

    return valid


def is_valid_arg_setting(sample: dict) -> bool:
    """Validates properly formatted arg setting

    :param sample: the potential blueprint arguments
    :type sample: dict
    :return: T: Valid argument setting
    :rtype: bool
    """
    valid = True
    if not isinstance(sample, dict):
        valid = False
    elif "argument_type" not in sample.keys():
        valid = False
    elif "default_value" not in sample.keys():
        valid = False
    elif "help_str" not in sample.keys():
        return False

    if not valid:
        return False

    if not isinstance(sample["argument_type"], ArgumentType):
        valid = False
    if not isinstance(sample["help_str"], str):
        valid = False
    if not isinstance(sample["argument_type"], ArgumentType):
        valid = False

    if not valid:
        return False

    return default_is_arg_type(sample)


def is_blueprint(sample: dict) -> bool:
    """verifies a dict is a blueprint

    :param sample: the potential blueprint
    :type sample: dict
    :return: T: is a blueprint
    :rtype: bool
    """
    valid = True

    try:
        if not isinstance(sample, dict):
            return False

        if "group_label" not in sample.keys() and valid:
            valid = False
        elif "type_label" not in sample.keys() and valid:
            valid = False

        if "arguments" not in sample.keys() and valid:
            valid = False
        else:
            if not isinstance(sample["arguments"], dict):
                valid = False

        if not valid:
            return False

        # value checks
        if not isinstance(sample["group_label"], str):
            valid = False
        if not isinstance(sample["type_label"], str):
            valid = False

        if not valid:
            return False

        for _, item in sample["arguments"].items():
            valid = is_valid_arg_setting(item)
            if not valid:
                return False

        return True
    except KeyError:
        # and to be extra safe.
        return False


def build_tree(
    blueprint: dict[str, Any],
    parent: anytree.Node
) -> None:
    """Uses Anytree to create a tree object for rendering. The function
    is designed to be recursive in order to be as flexible as possible.

    :param blueprint: The blueprint for the item
    :type blueprint: FactoryBlueprint
    :param parent: The parent
    :type parent: anytree.Node
    """
    if parent is None:
        parent = anytree.Node(blueprint["group_label"])

    item_node = anytree.Node(blueprint["type_label"], parent=parent)

    for key, item in blueprint["arguments"].items():
        please_add = True
        if isinstance(item, dict):
            if "arguments" in item and "group_label" in item \
                    and "type_label" in item:
                child_node = anytree.Node(key, parent=item_node)
                build_tree(blueprint=item, parent=child_node)
                please_add = False
        if isinstance(item, list):
            please_add = False
            list_node = anytree.Node(key, parent=item_node)
            for entry in item:
                if "arguments" in entry and "group_label" in entry \
                        and "type_label" in entry:
                    entry = cast(dict, entry)
                    build_tree(blueprint=entry, parent=list_node)
                else:
                    for item_entry in entry:
                        child_node = anytree.Node(item_entry, parent=list_node)
                        anytree.Node(f"{key}:{item}", parent=item_node)

        if please_add:
            if isinstance(item, str):
                anytree.Node(f"{key}:{item}", parent=item_node)
            elif isinstance(item, Number):
                anytree.Node(f"{key}:{item}", parent=item_node)
            elif isinstance(item, bool):
                anytree.Node(f"{key}:{item}", parent=item_node)
            elif isinstance(item, dict):
                anytree.Node(f"{key}:{item}", parent=item_node)
            elif isinstance(item, list):
                anytree.Node(f"{key}:{item}", parent=item_node)
            else:
                # catching unexpected configuration items.
                msg = f"Unable to build tree for item: {key}: {item}"
                logging.warning(msg)


def convert_exp_def_to_string(
        exp_def: Union[FactoryBlueprint, dict, ComponentDef]
) -> str:
    """Converts an experiment definition to a string

    :param exp_def: The experiment/blueprint
    :type blueprint: FactoryBlueprint
    :return: the tree as a string
    :rtype: str
    """
    root_node = anytree.Node(exp_def["group_label"])
    build_tree(blueprint=dict(exp_def), parent=root_node)
    return anytree.RenderTree(root_node, maxlevel=None).by_attr("name")
    # return anytree.RenderTree(root_node, maxlevel=5)
